@@ -16,22 +16,27 @@
 
 from persistent import Persistent
 from persistent.dict import PersistentDict
+import sqlalchemy
+from sqlalchemy.orm import scoped_session, sessionmaker, class_mapper
+import threading.local
 
 # import Zope3 interfaces
+from zope.security.interfaces import NoInteraction
 
 # import local interfaces
-from interfaces import IAlchemyEngineUtility
+from ztfy.alchemy.interfaces import IAlchemyEngineUtility
 
 # import Zope3 packages
-from zope.app.container.contained import Contained
+from zope.container.contained import Contained
 from zope.interface import implements
-from zope.component import getUtility
+from zope.component import adapter, getUtility
+from zope.publisher.interfaces import IEndRequestEvent
 from zope.schema.fieldproperty import FieldProperty
 
 # import local packages
-import sqlalchemy
-from sqlalchemy.orm import scoped_session, sessionmaker, class_mapper
-from datamanager import ZopeTransactionExtension, join_transaction, STATUS_READONLY
+from ztfy.alchemy.datamanager import ZopeTransactionExtension, join_transaction, \
+                                     _SESSION_STATE, STATUS_ACTIVE, STATUS_READONLY
+from ztfy.utils.request import getRequest, getRequestData, setRequestData
 
 
 class AlchemyEngineUtility(Persistent):
@@ -73,7 +78,7 @@ class AlchemyEngineUtility(Persistent):
             return engine
         kw = {}
         kw.update(self.kw)
-        self._v_engine = sqlalchemy.create_engine(self.dsn,
+        self._v_engine = sqlalchemy.create_engine(str(self.dsn),
                                                   echo=self.echo,
                                                   pool_size=self.pool_size,
                                                   pool_recycle=self.pool_recycle,
@@ -115,13 +120,53 @@ def getEngine(engine):
         engine = getUtility(IAlchemyEngineUtility, engine).getEngine()
     return engine
 
-def getSession(engine, join=True, status=STATUS_READONLY):
-    engine = getEngine(engine)
-    factory = scoped_session(sessionmaker(bind=engine, twophase=True, extension=ZopeTransactionExtension()))
-    result = factory()
-    if join:
-        join_transaction(result, initial_state=status)
-    return result
+
+class SessionsStorage(threading.local):
+    """Thread local sessions storage"""
+
+    engines = {}
+
+    def clear(self):
+        self.engines.clear()
+
+sessions = SessionsStorage()
+
+
+def getSession(engine, join=True, status=STATUS_ACTIVE, request=None):
+    """Get a new SQLAlchemy session
+    
+    Session is stored in request and in a sessions storage"""
+    session = None
+    if request is None:
+        try:
+            request = getRequest()
+        except NoInteraction:
+            pass
+    if request is not None:
+        session_data = getRequestData('ztfy.alchemy.session', request, {})
+        session = session_data.get(engine)
+    if session is None:
+        _engine = getEngine(engine)
+        if _engine in sessions.engines:
+            session = sessions.engines[_engine]
+            if status != STATUS_READONLY:
+                _SESSION_STATE[id(session)] = status
+        if session is None:
+            factory = scoped_session(sessionmaker(bind=_engine, twophase=True, extension=ZopeTransactionExtension()))
+            session = factory()
+            if join:
+                join_transaction(session, initial_state=status)
+            sessions.engines[engine] = session
+        if (request is not None) and (session is not None):
+            session_data[engine] = session
+            setRequestData('ztfy.alchemy.session', session_data, request)
+    return session
+
+
+@adapter(IEndRequestEvent)
+def handleEndRequest(event):
+    """Clear thread's sessions on request end"""
+    sessions.clear()
 
 
 class MetadataManager(object):
